/*
  ds_adc.s, version 1
  
  Source code of the Delta Sigma Analog to Digital Converter
  Written for entire Atmel AVR family by
  Uros Platise, uros.platise@ijs.si, MAPP (c) Copyright April 1999
  
  Notes: There should be a hardware NEGATIVE loopback.
         Tie reference voltage to + input and Uc to - input
	 of the analog comparator.
	
	 If the target device is AT90S1200, the _ds_adc 
	 function does not save the registers. It also clears 
	 global interrupt bit on call and enables global 
	 interrupt mask on return!
	
	 Compile with -A or -T switch.
	
	 If you use older version than ava-0.3.0, then
	 pass the numbers to the configuration macros and never macros.
	 See examples in the file "ds_adc.h".
*/
	seg flash.dsadc
	
#include "avr.inc"

#define __user_config__			define extern virtual

	/*--------------------------*/
	/* ADC User's Configuration */
	/*==========================*/

/* Compensate ADC offset voltage */
#__user_config__ DS_ADC_OFFSET_VOLTAGE	4
#__user_config__ DS_ADC_EXPIRE_PERIOD	0x3fff

/* Comparator Output; to be initialized before calling
   _ds_adc function. 
*/
#__user_config__ DS_ADC_OUT_PIN		2
#__user_config__ DS_ADC_OUT_PORT	0x18


/* 
   AVRGCC_RETURN_REGISTER
   
   For avr-gcc version < 0.8 this was r2 but
   later versions have changed it to r24. 
   Please set this macro according to your version of the 
   avr-gcc until standard file is provided by the avr-gcc for
   calling convention.
*/

;#define AVRGCC_RETURN_REGISTER		r2	/* ver < 0.8 */
#define AVRGCC_RETURN_REGISTER		r24	/* ver 0.8 */

/* Port Declarations */
#define DS_ADC_DD_PORT			(DS_ADC_OUT_PORT-1)
#define DS_ADC_IN_PORT			(DS_ADC_OUT_PORT-2)


	/*---------------------------*/
	/* ADC Declarations and Code */
	/*===========================*/

/* Registers: port output, debouncers "weight" and counters */
#define OutVal		r16
#define dbw		r17
#define num_one		r18
#define T_h		r31	/* index register ZH */
#define T_l		r30	/* ZL */
#define tr_h		r29	/* index register YH */
#define tr_l		r28	/* YL */


/*
  avr-gcc function prototype: 
  
  	void ds_adc_init();
*/
public _ds_adc_init:

#ifndef AT90S1200
	push	T_l
	push	T_h
#endif
	
	/* Initialize Comparator */	
	cbi	DS_ADC_OUT_PORT, DS_ADC_OUT_PIN
	sbi	DS_ADC_DD_PORT, DS_ADC_OUT_PIN
	
	/* Set Uc on the bottom end */
	ldi	T_h, high(DS_ADC_EXPIRE_PERIOD)
	ldi	T_l, low(DS_ADC_EXPIRE_PERIOD)
get_down:
#ifdef AT90S1200	
	subi	T_l,low(1)
	sbci	T_h,high(1)
#else
	sbiw	T_l,1	
#endif	
	brne	get_down_debnc
	rjmp	init_exit
	
get_down_debnc:
	sbis	ACSR, ACO
	rjmp	get_down
	sbis	ACSR, ACO
	rjmp	get_down
	sbis	ACSR, ACO
	rjmp	get_down
	sbis	ACSR, ACO	
	rjmp	get_down	
	sbis	ACSR, ACO
	rjmp	get_down
	sbis	ACSR, ACO
	rjmp	get_down
	sbis	ACSR, ACO
	rjmp	get_down
	sbis	ACSR, ACO	
	rjmp	get_down	
		
init_exit:
#ifndef AT90S1200
	pop	T_h
	pop	T_l
#endif	
	ret

/*
  avr-gcc function prototype: 
  
  	unsigned char ds_adc();
*/
public _ds_adc:

#ifndef AT90S1200
	push	OutVal
	in	OutVal, SREG
	push	OutVal		/* store current Interrupt Status */
	push	T_l
	push	T_h
	push	tr_l
	push	tr_h
	push	dbw
	push	num_one
#endif
	cli			/* disable interrupts for approx: 10 ms! */
		
	/* Set Uc on the top end */
	ldi	T_h, high(DS_ADC_EXPIRE_PERIOD)
	ldi	T_l, low(DS_ADC_EXPIRE_PERIOD)
	sbi	DS_ADC_OUT_PORT, DS_ADC_OUT_PIN		
get_up:
#ifdef AT90S1200	
	subi	T_l,low(1)
	sbci	T_h,high(1)
#else
	sbiw	T_l,1	
#endif	
	brne	get_up_debnc
	clr	tr_l
	rjmp	dsadc_exit
	
get_up_debnc:
	sbic	ACSR, ACO
	rjmp	get_up
	sbic	ACSR, ACO
	rjmp	get_up
	sbic	ACSR, ACO
	rjmp	get_up
	sbic	ACSR, ACO
	rjmp	get_up	
	sbic	ACSR, ACO
	rjmp	get_up
	sbic	ACSR, ACO
	rjmp	get_up
	sbic	ACSR, ACO
	rjmp	get_up
	sbic	ACSR, ACO
	rjmp	get_up
	
	/* init variables */
	ldi	num_one, 1
	ldi	T_h, 0x0F	/* cycle for 0xfff times */
	ser	T_l
	clr	tr_h
	clr	tr_l
	clr	dbw
					
	/* Get down and start conversion */
	cbi	DS_ADC_OUT_PORT, DS_ADC_OUT_PIN		
	
	/* prepare values and start conversion */
	in	OutVal, DS_ADC_OUT_PORT
	
	/* each pass: 19 cycles, dt = 2.375 us, symetric */
cycle_dsadc:

	/* count fall time */
	clc
	sbis	DS_ADC_IN_PORT, DS_ADC_OUT_PIN
	add	tr_l, num_one
	brcc	leave_tr_h
	inc	tr_h
leave_tr_h:	
	
	/* debounce ACO */
	lsl	dbw
	sbic	ACSR, ACO
	inc	dbw	
	cpi	dbw, 0			/* if dbw == low set Vol*/
	brne	_not_zero_
	cbr	OutVal, BV(DS_ADC_OUT_PIN)
_not_zero_:		
	cpi	dbw, 0xff		/* if dbw == high set Voh*/
	brne	_not_one_
	sbr	OutVal, BV(DS_ADC_OUT_PIN)
_not_one_:	

/* update output and loop again */
	out	DS_ADC_OUT_PORT, OutVal	
	
#ifdef AT90S1200	
	subi	T_l,low(1)
	sbci	T_h,high(1)
#else
	sbiw	T_l,1	
#endif	
	brne	cycle_dsadc
			
/* Divide by 16:
   swap lower nibble of the tr_h and append higher nibble of the tr_l 
*/
	swap	tr_h
	swap	tr_l
	andi	tr_h, 0xf0
	andi	tr_l, 0x0f
	or	tr_l, tr_h
	
/* compensate offset voltage error */
	subi	tr_l, DS_ADC_OFFSET_VOLTAGE
	cpi	tr_l, 0
	brmi	res_equal_zero

		
/* return 8 bit result */
dsadc_exit:		
	mov	AVRGCC_RETURN_REGISTER, tr_l
	
/* get the capacitor Uc down and thus prepare RCR for the next conversion */
	cbi	DS_ADC_OUT_PORT, DS_ADC_OUT_PIN	
	
#ifdef AT90S1200
	reti
#else
	pop	num_one
	pop	dbw
	pop	tr_h
	pop	tr_l
	pop	T_h
	pop	T_l
	pop	OutVal
	out	SREG, OutVal		/* restore old Interrupt Status */
	pop	OutVal
	ret
#endif

/* if x < 0 and x > (255-offset) then x = 0 */
res_equal_zero:
	cpi	tr_l, 0x100 - DS_ADC_OFFSET_VOLTAGE
	brmi	dsadc_exit
	clr	tr_l
	rjmp	dsadc_exit
	
